/*
 * MokaBots - a package including bots to learn robocoding
 * Copyright (C) 2003  Simone Chiaretta
 * 
 * This file is part of Agenzia Viaggi.  
 *
 * MokaBots is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * For more information about this program please refer to the README file.
 *
 */


package moka;
import robocode.*;
import java.awt.Color;


/**
 * MokaEnhanced - a robot by Simone Chiaretta <simone@piyosailing.com>
 */
public class AdvancedMokaRobot extends AdvancedRobot
{
	// Costante PI greco
	static final double PI = Math.PI;
	
	// Variabili usate per salvare la posizione del nemico
	double enemyPosX;
	double enemyPosY;
	double enemyBearing;
	
	boolean clockwise = true;
	
	//variabile statica per configurare il misuratore di prossimit
	static double allowedWallDistance=60;

		
	/**
	 * run: MokaEnhanced's default behavior
	 */
	public void run() {
		
		//Imposto il colore
		setColors(Color.red,Color.blue,Color.green);
		//Rendo indipendenti torretta dalla base
		setAdjustGunForRobotTurn(true);
		
		//Aggiungo evento per evitare di colpire un muro (e quindi di evitare di perdere punti)
		addCustomEvent(new Condition("NearWall") {
			public boolean test() {
				return (getX()<allowedWallDistance || getY()<allowedWallDistance || getX()>getBattleFieldWidth()-allowedWallDistance  || getY()>getBattleFieldHeight()-allowedWallDistance);
			}
		});
		//Centro la posizione del robot
		centerRobot();
		//Attendo che tutto lo spostamento sia completo (cio abbia raggiunto il centro)
		waitFor(new MoveCompleteCondition(this));
		while(true) {
			//attendo di aver completato la rotazione
			waitFor(new TurnCompleteCondition(this));
			//inverto la direzione di rotazione
			reverseDirection();
		}
	}

	/**
	 * onScannedRobot: What to do when you see another robot
	 */
	public void onScannedRobot(ScannedRobotEvent e) {
		//Calcolo la posizone del nemico
		calculatePosition(e.getDistance(),e.getBearingRadians());
		//Calcolo di quanto spostare la torretta per mirare al nemico
		double dest=Math.toDegrees(normaliseBearing(Math.toRadians(getGunHeading())-enemyBearing));
		//Giro torretta
		setTurnGunLeft(dest);
		//Sparo
		fire(2);
	}


	/**
	 * onCustomEvent: what to do when a custom event occurs
	 */
	public void onCustomEvent(CustomEvent ev) {
		Condition cd = ev.getCondition();
		System.out.println("event with " + cd.getName());
		if (cd.getName().equals("NearWall")) {
			centerRobot();
		}
	}
	

	/**
	 * onHitByBullet: What to do when you're hit by a bullet
	 */
	public void onHitByBullet(HitByBulletEvent e) {
		reverseDirection();
	}

	/**
	 * onHitByBullet: What to do when you're hit by a bullet
	 */
	public void onHitWall(HitWallEvent e) {
		centerRobot();
	}

	/**
	 * onHitRobot: What to do when you hit a robot
	 */	
	public void onHitRobot(HitRobotEvent e) {
		reverseDirection();
	}
	
	/**
	 * Calcolo la posizone assoluta del robot
	 */
	public void calculatePosition(double distance, double bearing) {
		enemyBearing=bearing+Math.toRadians(getHeading());
	}

	/**
	 * Inverto al direzione dello spostamento
	 */	
	protected void reverseDirection() {
		if(clockwise) {
			setTurnRightRadians(PI);
			setAhead(2000);
		} else {
			setTurnLeftRadians(PI);
			setBack(2000);
		}
		setTurnGunRightRadians(PI);
		clockwise=!clockwise;
		System.out.println("inversione");
		
	}
	
	/**
	 * Muovo il robot al centro del campo di gioco
	 */		
	protected void centerRobot() {
		goTo(getBattleFieldWidth()/2,getBattleFieldHeight()/2);
		waitFor(new TurnCompleteCondition(this));
	}
	
/****************************************************************************************************
 ************************ Funzioni di utilit riusabili *********************************************
 ****************************************************************************************************/
 
	/**
	 * Mi assicuro che l'angolo sia tra -180 e 180
	 */	
	protected double normaliseBearing(double ang) {
		if (ang > PI)
			ang -= 2*PI;
		if (ang < -PI)
			ang += 2*PI;
		return ang;
	}
	
	/**
	 * Mi assicuro che l'angolo sia tra 0 e 360
	 */		
	protected double normaliseHeading(double ang) {
		if (ang > 2*PI)
			ang -= 2*PI;
		if (ang < 0)
			ang += 2*PI;
		return ang;
	}
	
	/**
	 * Calcola la distanza tra due punti
	 */
	protected double getRange (double x1, double y1, double x2, double y2){
	
		double dx=x2-x1;
		double dy=y2-y1;
		
		double h=Math.sqrt(dx*dx + dy*dy);
		return h;
	}
	
	/**
	 * Calcola l'angolo tra due punti
	 */
	protected double absBearing(double x1,double y1, double x2,double y2 )
	{
		double dx=x2-x1;
		double dy=y2-y1;

		return Math.atan2(dx,dy);
	}	
	
	/**
	 * Mi sposto verso le coordinate x e y
	 */
	protected void goTo(double x, double y) {
		System.out.println("x="+x+",y="+y);
	    double dist = getRange(getX(),getY(),x,y);
	    double angle = absBearing(getX(),getY(),x,y);
	    double r = turnTo(angle);
	    setAhead(dist * r);
	}


	/**
	 * Curva del minimo angolo per raggiungere un direzione e ritorna il verso dello spostamento necessario
	 */
	protected double turnTo(double angle) {
	    double turnAmount;
    	double dir=1.0;
	    turnAmount = normaliseBearing(angle - getHeadingRadians());
	    if (turnAmount > PI/2) {
	        turnAmount -= PI;
	        dir = -1.0;
	    }
	    else if (turnAmount < -PI/2) {
	        turnAmount += PI;
	        dir = -1.0;
	    }
	    else {
	        dir = 1.0;
	    }
	    setTurnRightRadians(turnAmount);
	    return dir;
	}	
	
}
